/************************************************************************
*	Module		:	I2C (slave) controller								*
*	Filename	:	i2c_slave.c											*
*																		*
*	NOTE		:	SSPIF must be polled in main.						*
************************************************************************/

#include "define.h"

//global variables
ULONG I2C_TIMEOUT;										//I2C module timeout
UCHAR I2C_STATE;										//I2C hardware status
bit I2C_IDLE;											//flag for MSSP state

//local global variables
UCHAR I2C_DATA_IN[I2C_DATA_SIZE];						//input buffer
UCHAR I2C_DATA_IN_POS;									//position input buffer
UCHAR I2C_DATA_OUT[I2C_DATA_SIZE];						//output buffer

//function prototypes
void I2CSlaveInit(void);
void I2CSlaveExec(void);
void _I2CWrite(UCHAR);
UCHAR _PackByte(UCHAR);
UCHAR _UnpackByte(UCHAR);

//***** Start of Code ***************************************************
void I2CSlaveInit(void)
	{
	I2C_IDLE = 1;										//default flags state
	TRISC3 = 1;	TRISC4 = 1;								//RC3 and RC4 as input
	
	SSPSTAT = 0x00;										//100KHz clock
	SSPADD = I2C_ADDRESS;								//address for this MCU
	SSPCON = 0x36;										//I2C slave mode (7-bit address)
	SSPCON2 = 0x01;										//I2C module is idle, clock stretching enabled
	
	SSPIF = 0;											//interruptnot used! SSPIF polled in main
	}

void I2CSlaveExec(void)
	{
	UCHAR temp;
	static UINT temp_i;

	CKP = 0;											//hold SCL
	SSPIF = 0;											//reset interrupt flag
	I2C_STATE = SSPSTAT;

	I2C_IDLE = 0;										//I2C module not idle
	I2C_TIMEOUT = MS_COUNTER + I2C_TIMEOUT_VALUE;		//setup timeout

	if((I2C_STATE & 0x2D) == 0x09)
		{												//master address write, state 1 (S=1 RW=0 DA=0 BF=1)
		temp = SSPBUF;									//dummy read
		I2C_DATA_IN_POS = 0;							//reset pointer

		if(SSPOV == 1) temp = SSPBUF; SSPOV = 0;
		}
	else if((I2C_STATE & 0x2D) == 0x29)
		{												//master data write, state 2 (S=1 RW=0 DA=1 BF=1)
		I2C_DATA_IN[I2C_DATA_IN_POS] = SSPBUF;			//slave data read
		I2C_DATA_IN_POS++;								//increment input buffer pointer

		if(I2C_DATA_IN_POS > 2)
			{
			switch(I2C_DATA_IN[0])
				{
				case I2C_CMD_DIO_DIR_WRITE: {			//write DIO direction
					TRISB = I2C_DATA_IN[1];
					I2C_DATA_IN[2] = _UnpackByte(I2C_DATA_IN[2]) | 0x18;
					TRISC &= 0x18;
					TRISC |= I2C_DATA_IN[2];
					break;
					}
				case I2C_CMD_DIO_WRITE: {				//writing DIO
					PORTB = I2C_DATA_IN[1];
					I2C_DATA_IN[2] = _UnpackByte(I2C_DATA_IN[2]) | 0x18;
					PORTC = (PORTC & 0x18) | I2C_DATA_IN[2];
					break;
					}
				case I2C_CMD_DIO_SET: {					//setting one DIO pin
					switch(I2C_DATA_IN[1])
						{
						case 0:  DIO0  = 1; break;
						case 1:  DIO1  = 1; break;
						case 2:  DIO2  = 1; break;
						case 3:  DIO3  = 1; break;
						case 4:  DIO4  = 1; break;
						case 5:  DIO5  = 1; break;
						case 6:  DIO6  = 1; break;
						case 7:  DIO7  = 1; break;
						case 8:  DIO8  = 1; break;
						case 9:  DIO9  = 1; break;
						case 10: DIO10 = 1; break;
						case 11: DIO11 = 1; break;
						case 12: DIO12 = 1; break;
						case 13: DIO13 = 1; break;
						}
					break;
					}
				case I2C_CMD_DIO_RESET: {				//resetting one DIO pin
					switch(I2C_DATA_IN[1])
						{
						case 0:  DIO0  = 0; break;
						case 1:  DIO1  = 0; break;
						case 2:  DIO2  = 0; break;
						case 3:  DIO3  = 0; break;
						case 4:  DIO4  = 0; break;
						case 5:  DIO5  = 0; break;
						case 6:  DIO6  = 0; break;
						case 7:  DIO7  = 0; break;
						case 8:  DIO8  = 0; break;
						case 9:  DIO9  = 0; break;
						case 10: DIO10 = 0; break;
						case 11: DIO11 = 0; break;
						case 12: DIO12 = 0; break;
						case 13: DIO13 = 0; break;
						}
					break;
					}
				}
			I2C_IDLE = 1;
			}

		if(SSPOV == 1) temp = SSPBUF; SSPOV = 0;
		}
	else if((I2C_STATE & 0x2D) == 0x0c)
		{												//master address read, state 3 (S=1 RW=1 DA=0 BF=0)
		temp = SSPBUF;									//read address to clear SSPBUF 

		switch(I2C_DATA_IN[0])
			{											//sending to master 2nd byte (li byte)
			case I2C_CMD_ANA_READ_0: 
			case I2C_CMD_ANA_READ_1:
			case I2C_CMD_ANA_READ_2:
			case I2C_CMD_ANA_READ_3: {					//read ANA
				temp_i = ADCRead(I2C_DATA_IN[0]);
				I2C_DATA_OUT[0] = temp_i;
				I2C_DATA_OUT[1] = temp_i >> 8;
				break;
				}
			case I2C_CMD_DIO_DIR_READ: {				//reading DIO direction
				I2C_DATA_OUT[0] = TRISB;
				I2C_DATA_OUT[1] = _PackByte(TRISC);
				break;
				}
			case I2C_CMD_DIO_READ: {					//reading DIO
				I2C_DATA_OUT[0] = PORTB;
				I2C_DATA_OUT[1] = _PackByte(PORTC);
				break;
				}
			default: {									//unknown command (send 0xffff)
				I2C_DATA_OUT[0] = 0xff;
				I2C_DATA_OUT[1] = 0xff;
				break;
				}
			}

		_I2CWrite(I2C_DATA_OUT[0]);
		}
	else if((I2C_STATE & 0x2D) == 0x2C)
		{												//master data read, state 4 (S=1 RW=1 DA=1 BF=0)
		_I2CWrite(I2C_DATA_OUT[1]);
		}
	else if((I2C_STATE & 0x2D) == 0x28)
		{												//master NACK (S=1 RW=0 DA=1 BF=0)
		I2C_IDLE = 1;									//end of good transmission
		}
	else
		{
		SSPEN = 0; SSPEN = 1;							//unknown mode - reset MSSP
		I2C_IDLE = 1;
		}
	CKP = 1;											//release SCL
	}

void _I2CWrite(UCHAR value)
	{
	do {} while((STAT_BF == 1) && (I2C_IDLE == 0));		//loop until buffer empty or timeout
	WCOL = 0;
	SSPBUF = value;										//write I2C value
	}

UCHAR _PackByte(UCHAR input)
	{
	UCHAR temp;

	temp = (input >> 2) & 0x38;
	return temp | (input & 0x07);
	}

UCHAR _UnpackByte(UCHAR input)
	{
	UCHAR temp;

	temp = (input << 2) & 0xe0;
	return temp | (input & 0x07);
	}

//***** End of Code *****************************************************
